@@ -25,7 +25,7 @@
 
 
 __all__ = ('Connector', 'RPCSession', 'MessageSession', 'Server', 'ExcessiveSessionCostError',
-           'BatchError', 'Concurrency')
+           'BatchError', 'Concurrency', 'ReplyAndDisconnect')
 
 
 import asyncio
@@ -39,15 +39,22 @@ from aiorpcx.curio import (
     timeout_after, spawn_sync, ignore_after, sleep
 )
 from aiorpcx.framing import (
-    NewlineFramer, BitcoinFramer,
-    BadMagicError, BadChecksumError, OversizedPayloadError
+    NewlineFramer, BitcoinFramer, BadMagicError, BadChecksumError, OversizedPayloadError
 )
 from aiorpcx.jsonrpc import (
-    Request, Batch, Notification, ProtocolError, RPCError, FinalRPCError,
+    Request, Batch, Notification, ProtocolError, RPCError,
     JSONRPC, JSONRPCv2, JSONRPCConnection
 )
 
 
+class ReplyAndDisconnect(Exception):
+    '''Force a session disconnect after sending result (a Python object or an RPCError).
+    '''
+
+    def __init__(self, result):
+        self.result = result
+
+
 class Connector(object):
 
     def __init__(self, session_factory, host=None, port=None, proxy=None,
@@ -212,10 +219,6 @@ class SessionBase(asyncio.Protocol):
         self.errors += 1
         self.bump_cost(self.error_base_cost + getattr(exception, 'cost', 0.0))
 
-    def _close(self):
-        if self.transport:
-            self.transport.close()
-
     # asyncio framework
     def data_received(self, framed_message):
         '''Called by asyncio when a message comes in.'''
@@ -257,13 +260,16 @@ class SessionBase(asyncio.Protocol):
         '''Called by asyncio when the connection closes.
 
         Tear down things done in connection_made.'''
-        self._address = None
-        self.transport = None
-        self.closed_event.set()
-        self._task.cancel()
-
-        # Release waiting tasks
-        self._can_send.set()
+        # Work around uvloop bug; see https://github.com/MagicStack/uvloop/issues/246
+        if self.transport:
+            self._address = None
+            self.transport = None
+            self.closed_event.set()
+            # Release waiting tasks
+            self._can_send.set()
+            # Cancelling directly leads to self-cancellation problems for member
+            # functions await-ing self.close()
+            self.loop.call_soon(self._task.cancel)
 
     # External API
     def is_send_buffer_full(self):
@@ -350,14 +356,14 @@ class SessionBase(asyncio.Protocol):
 
     async def close(self, *, force_after=30):
         '''Close the connection and return when closed.'''
-        self._close()
-        if self._task:
-            async with ignore_after(force_after):
+        if self.transport:
+            self.transport.close()
+            try:
+                async with ignore_after(force_after):
+                    await self.closed_event.wait()
+            except TaskTimeout:
+                self.abort()
                 await self.closed_event.wait()
-            self.abort()
-            await self.closed_event.wait()
-            with suppress(CancelledError):
-                await self._task
 
 
 class MessageSession(SessionBase):
@@ -378,7 +384,7 @@ class MessageSession(SessionBase):
                     f'disconnecting'
                 )
                 self._bump_errors(e)
-                self._close()
+                await self.close()
             except OversizedPayloadError as e:
                 command, payload_len = e.args
                 self.logger.error(
@@ -386,7 +392,7 @@ class MessageSession(SessionBase):
                     f'{command}, disconnecting'
                 )
                 self._bump_errors(e)
-                self._close()
+                await self.close()
             except BadChecksumError as e:
                 payload_checksum, claimed_checksum = e.args
                 self.logger.warning(
@@ -415,7 +421,7 @@ class MessageSession(SessionBase):
             self.logger.info(f'incoming request timed out after {timeout} secs')
             self._bump_errors()
         except ExcessiveSessionCostError:
-            self._close()
+            await self.close()
         except CancelledError:
             raise
         except Exception:
@@ -563,6 +569,7 @@ class RPCSession(SessionBase):
 
     async def _throttled_request(self, request):
         '''Process a single request, respecting the concurrency limit.'''
+        disconnect = False
         try:
             timeout = self.processing_timeout
             async with timeout_after(timeout):
@@ -575,8 +582,12 @@ class RPCSession(SessionBase):
         except TaskTimeout:
             self.logger.info(f'incoming request {request} timed out after {timeout} secs')
             result = RPCError(JSONRPC.SERVER_BUSY, 'server busy - request timed out')
+        except ReplyAndDisconnect as e:
+            result = e.result
+            disconnect = True
         except ExcessiveSessionCostError:
-            result = FinalRPCError(JSONRPC.EXCESSIVE_RESOURCE_USAGE, 'excessive resource usage')
+            result = RPCError(JSONRPC.EXCESSIVE_RESOURCE_USAGE, 'excessive resource usage')
+            disconnect = True
         except CancelledError:
             raise
         except Exception:
@@ -589,9 +600,8 @@ class RPCSession(SessionBase):
                 await self._send_message(message)
         if isinstance(result, Exception):
             self._bump_errors(result)
-            if isinstance(result, FinalRPCError):
-                # Don't await self.close() because that is self-cancelling
-                self._close()
+        if disconnect:
+            await self.close()
 
     async def _send_concurrent(self, message, event, request_count):
         async with self._outgoing_concurrency:
