@@ -25,12 +25,14 @@
 
 '''SOCKS proxying.'''
 
+import asyncio
 import collections
-import ipaddress
+from ipaddress import IPv4Address, IPv6Address
+import socket
 import struct
 from functools import partial
 
-from aiorpcx.socket import getaddrinfo, SOCK_STREAM, socket as wrapped_socket
+from .util import classify_host
 
 
 __all__ = ('SOCKSUserAuth', 'SOCKS4', 'SOCKS4a', 'SOCKS5', 'SOCKSProxy',
@@ -58,6 +60,7 @@ class NeedData(Exception):
 
 
 class SOCKSBase(object):
+    '''Stateful as written so good for a single connection only.'''
 
     @classmethod
     def name(cls):
@@ -102,18 +105,15 @@ class SOCKS4(SOCKSBase):
 
     @classmethod
     def _check_host(cls, host):
-        if not isinstance(host, ipaddress.IPv4Address):
-            try:
-                host = ipaddress.IPv4Address(host)
-            except ValueError:
-                raise SOCKSProtocolError(
-                    f'SOCKS4 requires an IPv4 address: {host}') from None
-        return host
+        result = classify_host(host)
+        if not isinstance(result, IPv4Address):
+            raise SOCKSProtocolError(f'SOCKS4 requires an IPv4 address: {host}')
+        return result
 
     def _start(self):
         self._state = self._first_response
 
-        if isinstance(self._dst_host, ipaddress.IPv4Address):
+        if isinstance(self._dst_host, IPv4Address):
             # SOCKS4
             dst_ip_packed = self._dst_host.packed
             host_bytes = b''
@@ -151,10 +151,10 @@ class SOCKS4a(SOCKS4):
 
     @classmethod
     def _check_host(cls, host):
-        if not isinstance(host, (str, ipaddress.IPv4Address)):
-            raise SOCKSProtocolError(
-                f'SOCKS4a requires an IPv4 address or host name: {host}')
-        return host
+        result = classify_host(host)
+        if not isinstance(result, (str, IPv4Address)):
+            raise SOCKSProtocolError(f'SOCKS4a requires an IPv4 address or host name: {host}')
+        return result
 
 
 class SOCKS5(SOCKSBase):
@@ -178,9 +178,9 @@ class SOCKS5(SOCKSBase):
         self._auth_bytes, self._auth_methods = self._authentication(auth)
 
     def _destination_bytes(self, host, port):
-        if isinstance(host, ipaddress.IPv4Address):
+        if isinstance(host, IPv4Address):
             addr_bytes = b'\1' + host.packed
-        elif isinstance(host, ipaddress.IPv6Address):
+        elif isinstance(host, IPv6Address):
             addr_bytes = b'\4' + host.packed
         elif isinstance(host, str):
             host = host.encode()
@@ -285,7 +285,7 @@ class SOCKSProxy(object):
         auth = 'username' if self.auth else 'none'
         return f'{self.protocol.name()} proxy at {self.address}, auth: {auth}'
 
-    async def _handshake(self, client, socket):
+    async def _handshake(self, client, sock, loop):
         while True:
             count = 0
             try:
@@ -295,43 +295,55 @@ class SOCKSProxy(object):
             else:
                 if message is None:
                     return
-                await socket.sendall(message)
+                await loop.sock_sendall(sock, message)
 
             if count:
-                data = await socket.recv(count)
+                data = await loop.sock_recv(sock, count)
                 if not data:
                     raise SOCKSProtocolError("EOF received")
                 client.receive_data(data)
 
     async def _connect_one(self, host, port):
-        '''Connect to the proxy and perform a handshake requesting a connection to (host, port).
+        '''Connect to the proxy and perform a handshake requesting a
+        connection to (host, port).
 
-        Return the open socket.
+        Return the open socket on success, or the exception on failure.
         '''
-        client = self.protocol(host, port, self.auth)
-        socket = wrapped_socket()
-        # Don't close the socket on error - see https://github.com/kyuupichan/aiorpcX/issues/8
-        await socket.connect(self.address)
-        await self._handshake(client, socket)
-        self.peername = socket.getpeername()
-        return socket
+        loop = asyncio.get_event_loop()
+
+        proxy_host, proxy_port = self.address[:2]
+        for info in await loop.getaddrinfo(proxy_host, proxy_port, type=socket.SOCK_STREAM):
+            # This object has state so is only good for one connection
+            client = self.protocol(host, port, self.auth)
+            sock = socket.socket(family=info[0])
+            try:
+                # A non-blocking socket is required by loop socket methods
+                sock.setblocking(False)
+                await loop.sock_connect(sock, info[4])
+                await self._handshake(client, sock, loop)
+                self.peername = sock.getpeername()
+                return sock
+            except (OSError, SOCKSProtocolError) as e:
+                exception = e
+                # Don't close the socket because of an asyncio bug
+                # see https://github.com/kyuupichan/aiorpcX/issues/8
+        return exception
 
     async def _connect(self, addresses):
-        '''For each address in addresses, connect to the proxy serially and perform a handshake
-        requesting a connection to the address.
+        '''Connect to the proxy and perform a handshake requesting a
+        connection to each address in addresses.
 
-        Return an (open_socket, address) pair on the first success.
+        Return an (open_socket, address) pair on success.
         '''
         assert addresses
 
         exceptions = []
         for address in addresses:
             host, port = address[:2]
-            try:
-                socket = await self._connect_one(host, port)
-                return socket, address
-            except Exception as exc:
-                exceptions.append(exc)
+            sock = await self._connect_one(host, port)
+            if isinstance(sock, socket.socket):
+                return sock, address
+            exceptions.append(sock)
 
         strings = set(f'{exc!r}' for exc in exceptions)
         raise (exceptions[0] if len(strings) == 1 else
@@ -344,27 +356,25 @@ class SOCKSProxy(object):
         if self.protocol is SOCKS4a:
             host, port = 'www.apple.com', 80
         else:
-            host, port = ipaddress.IPv4Address('8.8.8.8'), 53
+            host, port = IPv4Address('8.8.8.8'), 53
 
-        try:
-            socket = await self._connect_one(host, port)
-        except Exception as exc:
-            # SOCKSFailure indicates something failed, but that we are
-            # likely talking to a proxy
-            return isinstance(exc, SOCKSFailure)
-        else:
-            await socket.close()
+        sock = await self._connect_one(host, port)
+        if isinstance(sock, socket.socket):
+            sock.close()
             return True
 
+        # SOCKSFailure indicates something failed, but that we are
+        # likely talking to a proxy
+        return isinstance(sock, SOCKSFailure)
+
     @classmethod
     async def auto_detect_address(cls, address, auth):
-        '''Try to detect a SOCKS proxy at address using the authentication
-        method (or None).  SOCKS5, SOCKS4a and SOCKS are tried in
-        order.  If a SOCKS proxy is detected a SOCKSProxy object is
-        returned.
+        '''Try to detect a SOCKS proxy at address using the authentication method (or None).
+        SOCKS5, SOCKS4a and SOCKS are tried in order.  If a SOCKS proxy is detected a
+        SOCKSProxy object is returned.
 
-        Returning a SOCKSProxy does not mean it is functioning - for
-        example, it may have no network connectivity.
+        Returning a SOCKSProxy does not mean it is functioning - for example, it may have
+        no network connectivity.
 
         If no proxy is detected return None.
         '''
@@ -378,12 +388,8 @@ class SOCKSProxy(object):
     async def auto_detect_host(cls, host, ports, auth):
         '''Try to detect a SOCKS proxy on a host on one of the ports.
 
-        Calls auto_detect for the ports in order.  Returns SOCKS are
-        tried in order; a SOCKSProxy object for the first detected
-        proxy is returned.
-
-        Returning a SOCKSProxy does not mean it is functioning - for
-        example, it may have no network connectivity.
+        Calls auto_detect_address for the ports in order.  Returning a SOCKSProxy does not
+        mean it is functioning - for example, it may have no network connectivity.
 
         If no proxy is detected return None.
         '''
@@ -400,27 +406,32 @@ class SOCKSProxy(object):
                                 family=0, proto=0, flags=0):
         '''Set up a connection to (host, port) through the proxy.
 
-        If resolve is True then host is resolved locally with getaddrinfo using family,
-        proto and flags, otherwise the proxy is asked to resolve host.
+        If resolve is True then host is resolved locally with
+        getaddrinfo using family, proto and flags, otherwise the proxy
+        is asked to resolve host.
 
-        The function signature is similar to loop.create_connection() with the same
-        result.  The attribute _address is set on the protocol to the address of the
-        successful remote connection.  Additionally raises SOCKSError if something goes
-        wrong with the proxy handshake.
+        The function signature is similar to loop.create_connection()
+        with the same result.  The attribute _address is set on the
+        protocol to the address of the successful remote connection.
+        Additionally raises SOCKSError if something goes wrong with
+        the proxy handshake.
         '''
+        loop = asyncio.get_event_loop()
         if resolve:
-            infos = await getaddrinfo(host, port, family=family, type=SOCK_STREAM,
-                                      proto=proto, flags=flags)
+            infos = await loop.getaddrinfo(host, port, family=family,
+                                           type=socket.SOCK_STREAM,
+                                           proto=proto, flags=flags)
             addresses = [info[4] for info in infos]
         else:
             addresses = [(host, port)]
 
-        socket, address = await self._connect(addresses)
+        sock, address = await self._connect(addresses)
 
         def set_address():
             protocol = protocol_factory()
             protocol._address = address
             return protocol
 
-        return await socket.create_connection(set_address, ssl=ssl,
-                                              server_hostname=host if ssl else None)
+        return await loop.create_connection(
+            set_address, sock=sock, ssl=ssl,
+            server_hostname=host if ssl else None)
